use base64::prelude::{Engine, BASE64_URL_SAFE_NO_PAD};
use eyre::{Context, Report, Result};
use foxchat::FoxError;
use rand::RngCore;
use sqlx::{PgExecutor, Pool, Postgres};

use crate::model::account::{Account, Role};

pub async fn get_user_by_username_and_password(
    pool: Pool<Postgres>,
    username: String,
    password: String,
) -> Result<Account> {
    let account = sqlx::query_as!(
        Account,
        r#"select
        id, username, email, password, role as "role: Role", avatar
        from accounts where username = $1"#,
        username
    )
    .fetch_one(&pool)
    .await
    .map_err(|e| -> Report {
        match e {
            sqlx::Error::RowNotFound => FoxError::NotFound.into(),
            _ => e.into(),
        }
    })?;

    if bcrypt::verify(password, &account.password).map_err(|_| FoxError::NotFound)? {
        return Ok(account);
    }

    Err(FoxError::NotFound.into())
}

pub async fn check_token(pool: &Pool<Postgres>, token: String) -> Result<Account> {
    let hash = sha256::digest(token);

    let account = sqlx::query_as!(
        Account,
        r#"select
    a.id, a.username, a.email, a.password, a.role as "role: Role", a.avatar
    from accounts a
    join tokens t on t.account_id = a.id
    where t.account_id = $1"#,
        hash
    )
    .fetch_optional(pool)
    .await?;

    match account {
        Some(a) => Ok(a),
        None => Err(FoxError::NotFound.into()),
    }
}

pub async fn create_token(executor: impl PgExecutor<'_>, user_id: &String) -> Result<String> {
    let token = generate_token_string();
    let hash = sha256::digest(&token);
    sqlx::query!(
        "insert into tokens (token, account_id) values ($1, $2)",
        hash,
        user_id
    )
    .execute(executor)
    .await?;

    Ok(token)
}

fn generate_token_string() -> String {
    let mut data = [0u8; 32];
    rand::thread_rng().fill_bytes(&mut data);

    BASE64_URL_SAFE_NO_PAD.encode(data)
}

const PASSWORD_COST: u32 = 12;

pub fn hash_password(password: String) -> Result<String> {
    bcrypt::hash(password, PASSWORD_COST).wrap_err("failed to hash password")
}
